"""
expression rendering
"""

from webview.dom.element import Element
import ast

import html
import json

from .dom import register, div, block, add_node, add, add_text, add_pre


def render(parent: Element, node: ast.expr):
    match type(node):
        case ast.BoolOp:
            return render_boolop(parent, node)
        case ast.NamedExpr:
            raise NotImplementedError('expression.render() not implemented for ast.NamedExpr')
        case ast.BinOp:
            return render_binop(parent, node)
        case ast.UnaryOp:
            return render_unaryop(parent, node)
        case ast.Lambda:
            raise NotImplementedError('expression.render() not implemented for ast.Lambda')
        case ast.IfExp:
            return render_ifexp(parent, node)
        case ast.Dict:
            return render_dict(parent, node)
        case ast.Set:
            raise NotImplementedError('expression.render() not implemented for ast.Set')
        case ast.ListComp:
            return render_list_comprehension(parent, node)
        case ast.SetComp:
            raise NotImplementedError('expression.render() not implemented for ast.SetComp')
        case ast.DictComp:
            raise NotImplementedError('expression.render() not implemented for ast.DictComp')
        case ast.GeneratorExp:
            raise NotImplementedError('expression.render() not implemented for ast.GeneratorExp')
        case ast.Await:
            raise NotImplementedError('expression.render() not implemented for ast.Await')
        case ast.Yield:
            raise NotImplementedError('expression.render() not implemented for ast.Yield')
        case ast.YieldFrom:
            raise NotImplementedError('expression.render() not implemented for ast.YieldFrom')
        case ast.Compare:
            return render_compare(parent, node)
        case ast.Call:
            return render_call(parent, node)
        case ast.FormattedValue:
            return render_formatted_value(parent, node)
        # case ast.Interpolation:
        #     raise NotImplementedError('expression.render() not implemented for ast.Interpolation')
        case ast.JoinedStr:
            return render_joinedstr(parent, node)
        # case ast.TemplateStr:
        #     raise NotImplementedError('expression.render() not implemented for ast.TemplateStr')
        case ast.Constant:
            return render_constant(parent, node)
        case ast.Attribute:
            return render_attribute(parent, node)
        case ast.Subscript:
            return render_subscript(parent, node)
        case ast.Starred:
            return render_starred(parent, node)
        case ast.Name:
            return render_name(parent, node)
        case ast.List:
            return render_list(parent, node)
        case ast.Tuple:
            return render_tuple(parent, node)
        case ast.Slice:
            return render_slice(parent, node)
        case default:
            raise ValueError(f"Unexpected ast expression type: {type(node)}")




#BoolOp(boolop op, expr* values)
#NamedExpr(expr target, expr value)
#BinOp(expr left, operator op, expr right)
#UnaryOp(unaryop op, expr operand)
#Lambda(arguments args, expr body)
#IfExp(expr test, expr body, expr orelse)
#Dict(expr?* keys, expr* values)
#Set(expr* elts)
#ListComp(expr elt, comprehension* generators)
#SetComp(expr elt, comprehension* generators)
#DictComp(expr key, expr value, comprehension* generators)
#GeneratorExp(expr elt, comprehension* generators)
## the grammar constrains where yield expressions can occur
#Await(expr value)
#Yield(expr? value)
#YieldFrom(expr value)
## need sequences for compare to distinguish between
## x < 4 < 3 and (x < 4) < 3
#Compare(expr left, cmpop* ops, expr* comparators)
#Call(expr func, expr* args, keyword* keywords)
#FormattedValue(expr value, int conversion, expr? format_spec)
#Interpolation(expr value, constant str, int conversion, expr? format_spec)
#JoinedStr(expr* values)
#TemplateStr(expr* values)
#Constant(constant value, string? kind)
#
## the following expression can appear in assignment context
#Attribute(expr value, identifier attr, expr_context ctx)
#Subscript(expr value, expr slice, expr_context ctx)
#Starred(expr value, expr_context ctx)
#Name(identifier id, expr_context ctx)
#List(expr* elts, expr_context ctx)
#Tuple(expr* elts, expr_context ctx)
#
#- can appear only in Subscript
#Slice(expr? lower, expr? upper, expr? step)
#
#

def render_boolop(parent: Element, node: ast.BoolOp) -> Element:
    elt = add_node(parent, node, "row gap")
    elt.classes.append(f"{read_boolop(node.op)}-sep")
    for v in node.values:
        # add(..., "row") is a wrapper required by the {op}-sep separator
        render(add(elt, "row gap"), v)
    return elt


"""
this code uses html data-attributes to encode the operator, and css to render it dynamically

the html output in theory:
<div class="operation" data-operator="+">
  <div class="operand">a</div>
  <div class="operand">b</div>
</div>

in practice, css support is lacking, so we need this:
<div class="operation" data-operator="+">
  <div class="operand" data-operator="+">a<div>
  <div class="operand">b</div>
</div>
"""
def render_binop(parent: Element, node: ast.BinOp) -> Element:
    elt = add_node(parent, node, "operation row gap")
    elt.attributes["data-operator"] = read_binaryop(node.op)
    operator_suffixed = add(elt, "row gap op-suffixed")
    # note: setting css variables from the DOM API is a pain
    # recent versions of pywebview do not have the same syntax for setting css properties

    # this code works in every version, but very ugly
    #webview.windows[1].evaluate_js(f"""
    #document.getElementById('{node.node_id}').style.setProperty('--operator', '"+"')
    #""")

    # this does NOT work in cpython with apt-packaged pywebview, but almost
    #operator_suffixed.style["--operator"] = f'\\"{read_binaryop(node.op)}\\"'

    # this works in pypy with pip-packaged pywebview, but pywebview is complicated to install without distro packaging
    #operator_suffixed.style["--operator"] = json.dumps(f'"{read_binaryop(node.op)}"')

    # the easy solution is to never set css variables from python:
    left = render(elt, node.left)
    operator_prefixed = add(elt, "row gap")
    operator_prefixed.attributes["data-operator"] = read_binaryop(node.op)
    right = render(operator_prefixed, node.right)
    #operator_suffixed.attributes["data-operator"] = read_binaryop(node.op)
    #left = render(operator_suffixed, node.left)
    #right = render(add(elt, "row gap"), node.right)
    return elt

def read_binaryop(op: ast.operator):
    # css classes cannot have special characters like +
    if isinstance(op, ast.Add):
        return "+"
    elif isinstance(op, ast.Sub):
        return "-"
    elif isinstance(op, ast.Mult):
        return "*"
    elif isinstance(op, ast.Div):
        pass
    elif isinstance(op, ast.FloorDiv):
        pass
    elif isinstance(op, ast.Mod):
        pass
    elif isinstance(op, ast.Pow):
        pass
    elif isinstance(op, ast.LShift):
        pass
    elif isinstance(op, ast.RShift):
        pass
    elif isinstance(op, ast.BitOr):
        return "|"
    elif isinstance(op, ast.BitXor):
        pass
    elif isinstance(op, ast.BitAnd):
        pass
    elif isinstance(op, ast.MatMult):
        pass
    raise NotImplementedError(f"unknown binary operator: {op}")


# TODO: go back to previous "operator separators" and replace them by DOM nodes
# (operators have semantic meaning, they are not just syntax)
# (keep the css for infix inlining if needed)
def render_unaryop(parent: Element, node: ast.UnaryOp) -> Element:
    elt = add_node(parent, node, "row")
    op = add(elt, text=read_unaryop(node.op))
    value = render(elt, node.operand)
    return elt


def read_unaryop(op: ast.unaryop):
    if isinstance(op, ast.USub):
        return "-"
    elif isinstance(op, ast.UAdd):
        return "+"
    elif isinstance(op, ast.Not):
        return "not "
    elif isinstance(op, ast.Invert):
        return "~"
    else:
        raise NotImplementedError("unknown unary operator")





def read_boolop(op: ast.boolop):
    if isinstance(op, ast.And):
        return "and"
    elif isinstance(op, ast.Or):
        return "or"
    else:
        raise NotImplementedError(f"unknown boolean operator: {op}")


def render_ifexp(parent: Element, node: ast.IfExp) -> Element:
    elt = add_node(parent, node, "row gap")
    condition = add(elt)
    test = render(condition, node.test)

    if_part = add(elt, "if-prefix row gap")
    if_expr = render(if_part, node.body)
    else_part = add(elt, "else-prefix row gap")
    else_expr = render(else_part, node.orelse)
    return elt


# TODO: test with more kinds of literals
def render_dict(parent: Element, node: ast.Dict):
    # TODO: multi-line rendering:
    # ... {
    #     ....
    # }
    # for now: render on a single line
    # TODO:
    elt = add_node(parent, node, "row braces")
    items = add(elt, "comma-sep")
    for i in range(len(node.keys)):
        k = node.keys[i]
        # TODO: None is for **d unpacking
        assert k is not None
        v = node.values[i]
        key = add(items, "row colon-suffix")
        expression.render(key)
        value = add(items)
        expression.render(value, v)


def render_list_comprehension(parent: Element, node: ast.ListComp) -> Element:
    elt = add_node(parent, node, "brackets row")
    spaced_content = add(elt, "row gap")
    selected = render(spaced_content, node.elt)
    # TODO: test with more than 1 generator
    generators = add(spaced_content, "row gap")
    for g in node.generators:
        expr = render_comprehension_generator(generators, g)
    return elt

def render_comprehension_generator(parent: Element, node: ast.comprehension) -> Element:
    # TODO: support async
    assert node.is_async == 0
    # TODO: support conditions in comprehensions
    assert len(node.ifs) == 0

    elt = add_node(parent, node, "for-prefix in-sep row gap")
    # need a wrapper (row gap is the interaction with the in suffix) for in-sep
    target = render(elt, node.target)
    it = render(add(elt, "row gap"), node.iter)
    return elt


def render_compare(parent: Element, node: ast.Compare) -> Element:
    # in python, comparisons can be complex sequences, like:
    # 1 < x < y < 6
    # 1 is called left
    # the operators are [<, <, <]
    # and the comparators are [x, y, 6]
    elt = add_node(parent, node, "compare row gap")
    left = render(elt, node.left)
    for op, cmp in zip(node.ops, node.comparators):
        add(elt, text=read_op(op))
        render(elt, cmp)
    return elt

def read_op(op: ast.operator):
    if isinstance(op, ast.Eq):
        return "=="
    elif isinstance(op, ast.NotEq):
        return "!="
    elif isinstance(op, ast.Lt):
        return "<"
    elif isinstance(op, ast.LtE):
        return "<="
    elif isinstance(op, ast.Gt):
        return ">"
    elif isinstance(op, ast.GtE):
        return ">="
    elif isinstance(op, ast.Is):
        return "is"
    elif isinstance(op, ast.IsNot):
        return "is not"
    elif isinstance(op, ast.In):
        return "in"
    elif isinstance(op, ast.NotIn):
        return "not in"
    else:
        raise NotImplementedError("unknown comparison operator")




# TODO: add more DOM encoding
def render_call(parent: Element, node: ast.Call) -> Element:
    elt = add_node(parent, node, "call row")
    func = render(elt, node.func)
    args = add(elt, "parens row")
    comma_separated = add(args, "comma-sep row")
    # positional args
    for arg in node.args:
        render(add(comma_separated, "row gap"), arg)
    # kwargs
    for kwarg in node.keywords:
        render_keyword_arg(add(comma_separated, "row gap"), kwarg)
        #eq_separated = add(comma_separated, "equal-sep row")
        #kw = add(eq_separated, text=kwarg.arg)
        #render(eq_separated, kwarg.value)
    return elt

# part of render_call(), also used by statement.render_class()
def render_keyword_arg(parent: Element, node: ast.keyword) -> Element:
    elt = add_node(parent, node, "equal-sep row")
    # TODO: use a wrapper for "row"
    # not needed here because there is only text,
    # but .equal-sep adds a ::before text content
    # and using a wrapper div would unify the DOM conventions with the rest of the code
    kw = add(elt, "row", text=node.arg)
    val = add(elt, "row")
    render(val, node.value)
    return elt

def render_formatted_value(parent: Element, node: ast.FormattedValue) -> Element:
    # TODO: support other conversion types:
    # -1: unspecified (default is str())
    # 97: !a, ascii
    # 114: !r, repr() output
    # 115: !s, str() output
    assert node.conversion == -1
    # can be a nested JoinedStr instead of None
    # TODO: support this attribute
    assert node.format_spec is None
    elt = add_node(parent, node)
    render(elt, node.value)
    return elt


# f"{x}<text>{y}"
# f-"-({(<x>)}-(<json-encoded text>)-({<y>}))-"
def render_joinedstr(parent: Element, node: ast.JoinedStr) -> Element:
    elt = add_node(parent, node, "row f-prefix")
    quoted = add(elt, "quotes row")
    for e in node.values:
        if isinstance(e, ast.FormattedValue):
            braced = add(quoted, "braces row")
            render(braced, e)
        else:
            assert isinstance(e, ast.Constant)
            assert isinstance(e.value, str)
            # quotes are replaced because pywebview runs f"elt.textContent = '{text}'"
            # which breaks if there is a single quote in the string
            # and needs a second layer of escaping for everything that is escaped
            # TODO: merge with literal string formatting
            text = json.dumps(e.value).replace("\\", "\\\\").replace("'", "\\'")
            # remove quotes (and let the css quotes surround the whole f-string)
            text = text[1:-1]
            #add(quoted, text=text)
            pre = add_pre(quoted, text=text)
            #parts.append(e.value)
    return elt


def render_constant(parent: Element, node: ast.Constant) -> Element:
    assert node.kind is None
    elt = add_node(parent, node, "literal")
    # TODO: test very carefully
    # string rendering is very fragile, due to pywebview doing a double eval with JS code execution
    # also make sure to check html escaping
    # and make sure that the code works in later versions with pypy
    # TODO: centralize string escaping to prevent debugging the same bugs in other places (f-strings, comparisons,...)

    # tools available:
    # -html.escape(), already done by pywebview
    # -json.dumps(), for JS quotes escaping (repr will not work)
    # -double escaping for the rest: .replace("\\", "\\\\")

    if isinstance(node.value, str):
        # TODO: use the DOM to encode constant types
        # TODO: render multiline ("\n") strings with the multiline syntax if they are top-level in the module

        # text = json.dumps(node.value)
        # pywebview uses JS eval to do things, which seems to break newline and quote escaping
        #text = json.dumps(node.value).replace("\\", "\\\\")
        # new fix imported from f-string (JoinedStr) tests
        text = json.dumps(node.value).replace("\\", "\\\\").replace("'", "\\'")
        elt.text = text
        # TODO: format multiline strings differently (""" """)
        # does not work:
        # (js eval replaces escaped "\n" by \n)
        #text = json.dumps(node.value).replace("'", "\\'")
        # remove quotes
        #text = text[1:-1]
        #if "\\n" in text:
        #    c = "triple-quotes"
        #else:
        #    c = "quotes"
        #pre = add_pre(elt, text)
        #pre.classes.append("bg-red")
        #elt.classes.append(c)
    else:
        elt.text = repr(node.value)
        #print(elt.text)
    return elt

def render_attribute(parent: Element, node: ast.Attribute) -> Element:
    elt = add_node(parent, node, "attribute row dot-sep")
    render(elt, node.value)
    add(add(elt, "row"), text=node.attr)
    return elt

def render_subscript(parent: Element, node: ast.Subscript) -> Element:
    # node.ctx is either ast.Load or ast.Store
    # Store if the subscript is in a left side of an assignment
    # Load if the subscript is in an expression to evaluate
    elt = add_node(parent, node, "row")
    indexed = render(elt, node.value)
    bracketed = add(elt, "brackets row")
    index = render(bracketed, node.slice)
    return elt


def render_starred(parent: Element, node: ast.Starred) -> Element:
    #print(node.ctx)
    elt = add_node(parent, node, "star-prefix row")
    render(elt, node.value)
    return elt

def render_name(parent: Element, node: ast.Name) -> Element:
    elt = add_node(parent, node, "symbol")
    elt.text = node.id
    return elt

def render_list(parent: Element, node: ast.List) -> Element:
    assert isinstance(node.ctx, ast.Load)
    elt = add_node(parent, node, "brackets row")
    comma_separated = add(elt, "comma-sep row")
    for x in node.elts:
        comma_ended = add(comma_separated, "row gap")
        render(comma_ended, x)
    return elt


def render_tuple(parent: Element, node: ast.Tuple) -> Element:
    #print(node.ctx)
    elt = add_node(parent, node, "parens row")
    if len(node.elts) == 0:
        pass
    elif len(node.elts) == 1:
        # display a single item with a comma after it
        comma_separated = add(elt, "comma-sep row")
        render(add(comma_separated, "row"), node.elts[0])
        # empty element for the comma
        add(comma_separated)
    else:
        comma_separated = add(elt, "comma-sep row")
        for x in node.elts:
            render(add(comma_separated, "row gap"), x)
    return elt


def render_slice(parent: Element, node: ast.Slice):
    elt = add_node(parent, node)
    colon_split = add(elt, "row colon-sep")
    left = add(colon_split, "row")
    if node.lower is not None:
        render(left, node.lower)
    center = add(colon_split, "row")
    if node.upper is not None:
        render(center, node.upper)
    if node.step is not None:
        right = add(colon_split, "row")
        render(right, node.step)
    return elt



"""
general DOM manipulation utilities
"""

from webview.dom.element import Element
from ast import AST

"""
(bidirectional) AST-DOM linking

AST nodes are linked to DOM elements with a shared unique id
"""


dom_mapping = {}
ast_mapping = {}

def make_counter():
    counter = 0
    def count():
        nonlocal counter
        counter += 1
        return counter
    return count

genid = make_counter()

def register(ast_node: AST, dom_element: Element):
    n = genid()
    dom_element.id = n
    # .id is used by some node types
    ast_node.node_id = n
    dom_mapping[n] = dom_element
    ast_mapping[n] = ast_node

"""
basic wrappers for html string formatting

In practice, with the new dynamic DOM renderer,
the html content is always empty (children are appended after creation)
"""

def div_old(html=""):
    return f"<div>{html}</div>"
def div():
    return "<div></div>"
# TODO: use for strings that contain spaces
def pre():
    return "<pre></pre>"

def block(html=""):
    return f"<div class='block'>{html}</div>"

# TODO: add an easy wrapper for elt = dom.create_element(div, parent=parent); elt.classes = [...]; register(elt)
# wrappers for Element creation, AST node registering and html class setup
def add(parent: Element, cls: str | list = None, text: str = None):
    elt = parent.append(div())
    if cls:
        elt.classes = cls.split(" ") if isinstance(cls, str) else cls
    if text:
        elt.text = text
    return elt

def add_pre(parent: Element, text: str):
    elt = parent.append(pre())
    elt.text = text
    return elt

# TODO: deprecated, use add() instead
def add_text(parent: Element, text: str):
    elt = add(parent)
    elt.text = text
    return elt

def add_node(parent: Element, node: AST, cls: str | list = None, text: str = None):
    elt = add(parent, cls, text)
    register(node, elt)
    return elt

# TODO: find an API to add children with wrapping divs

#def add_managed_item(parent: Element, )
    #def comma_separated_list(parent: Element):
    #    elt = add(parent, "comma-sep")
    #    return elt
    #
    #def add_item(lst, *args)

#def add(parent, html):
#    parent.append(html)

#def row():
#    return "<div class='row'></div>"




"""
This is the source code of the IDE you are running.

TODO: refactor html generation
TODO: write helpers for body generation
"""

# TODO: write tests
#print(1 < 3 < 5)

import webview
import html
import ast
import sys
import code
import os

# used to load css files in the python import path
from importlib import resources

#import multiprocessing
#import marshal
# force=True: do it even in the child process (otherwise, a RuntimeError is raised)
#multiprocessing.set_start_method("spawn", force=True)

#def run_code(encoded_bytecode, module_dict):
#    bytecode = marshal.loads(encoded_bytecode)
#    exec(bytecode, module_dict)



from types import ModuleType

# test import, needed for exhaustiveness
import untext.rendering.statement as statement, untext.rendering.expression as expression
def f(a, b):
    pass

from untext.rendering import dom
from untext.rendering.dom import div_old as div, block

#import rendering.statement, rendering.expression

from untext.rendering import statement, expression




"""
helpers to create DOM from an AST
"""
def render_module(module_node):
    # unknown value
    assert len(module_node.type_ignores) == 0
    children = ["<div>"]
    for elt in module_node.body:
        # top-level strings are usually multiline
        # TODO: refactor to make the logic more explicit
        if isinstance(elt, ast.Expr) and isinstance(elt.value, ast.Constant) and isinstance(elt.value.value, str):
            lines = elt.value.value.split("\n")
            lines[0] = '"""' + lines[0]
            lines[-1] += '"""'
            lines = [line if line else "<br>" for line in lines]
            lines = [f"<div>{line}</div>" for line in lines]
            children.append("".join(lines))
            # failed implementation
            #rendered_text = elt.value.value.replace("\n", "</div><div>")
            #print(rendered_text)
            #print("\n" in rendered_text)
            #children.append(f'<div>"""</div><div>{elt.value.value}</div><div>"""</div>')
        else:
            children.append(render_statement(elt))
    children.append("</div>")
    return "".join(children)


def render_statement(node):
    if isinstance(node, ast.Import):
        return render_import(node)
    elif isinstance(node, ast.ImportFrom):
        return render_importfrom(node)
    elif isinstance(node, ast.FunctionDef):
        return render_funcdef(node)
    elif isinstance(node, ast.Expr):
        return render_expr_statement(node)
    elif isinstance(node, ast.With):
        return render_with(node)
    elif isinstance(node, ast.Assign):
        return render_assign(node)
    elif isinstance(node, ast.While):
        return render_while(node)
    elif isinstance(node, ast.For):
        return render_for(node)
    elif isinstance(node, ast.Assert):
        return render_assert(node)
    elif isinstance(node, ast.Return):
        return render_return(node)
    elif isinstance(node, ast.If):
        return render_if(node)
    elif isinstance(node, ast.Pass):
        return render_pass(node)
    elif isinstance(node, ast.Try):
        return render_try(node)
    elif isinstance(node, ast.Raise):
        return render_raise(node)
    elif isinstance(node, ast.AugAssign):
        targ = render_expr(node.target)
        op = render_binaryop(node.op)
        val = render_expr(node.value)
        return f"{targ} {op}= {val}"
    elif isinstance(node, ast.ClassDef):
        # TODO:
        return ""
    elif isinstance(node, ast.Nonlocal):
        # TODO:
        return ""
    elif isinstance(node, ast.Match):
        # TODO:
        return ""
    else:
        raise NotImplementedError(type(node))


def render_import(import_node):
    # TODO: support more than 1 import per statement
    # use case: import (... as ..., ... as ...)
    #assert len(import_node.names) == 1
    #for alias in import_node.names:
    alias = import_node.names[0]
    if alias.asname is not None:
        return div(f"import {alias.name} as {alias.asname}")
    else:
        return div(f"import {alias.name}")

def render_importfrom(node):
    fragments = []
    for alias in node.names:
        if alias.asname is None:
            fragments.append(f"{alias.name}")
        else:
            fragments.append(f"{alias.name} as {alias.asname}")
    # TODO: support multiline syntax if the fragments are too long
    fragments = ", ".join(fragments)
    line = f"from {node.level * '.'}{node.module} import {fragments}"
    return div(line)



def render_funcdef(funcdef_node):
    n = funcdef_node
    body = []
    # type_params is a 3.12+ feature
    # TODO: use type_comment, wait for pypy to reach 3.12 or stop supporting pypy
    #assert len(n.type_params) == 0
    header = f"def {n.name}({render_arguments(n.args)}):"
    for statement in n.body:
        body.append(render_statement(statement))
    # TODO: support decorators in function definitions
    assert len(n.decorator_list) == 0
    # never seen otherwise yet
    # TODO: (done in the dynamic renderer)
    if n.returns is not None:
        pass
    assert n.type_comment is None

    header = div(header)
    body = "".join(body)
    body = block(body)
    return div(header + body)

def render_with(node):
    body = "".join([render_statement(statement) for statement in node.body])
    items = ", ".join([render_withitem(item) for item in node.items])
    assert node.type_comment is None
    header = div(f"with {items}:")
    body = block(body)
    result = div(header + body)
    return result

def render_withitem(node):
    expr = render_expr(node.context_expr)
    if node.optional_vars:
        name = render_expr(node.optional_vars)
        return f"{expr} as {name}"
    else:
        return expr


def render_assign(assign_node):
    targets = tuple([render_expr(t) for t in assign_node.targets])
    value = render_expr(assign_node.value)
    assert assign_node.type_comment is None
    if len(targets) == 1:
        target = targets[0]
        return div(f"{target} = {value}")
    else:
        return div(f"{tuple(targets)} = {value}")

def render_while(node):
    test = render_expr(node.test)
    body = [render_statement(statement) for statement in node.body]
    else_body = [render_statement(statement) for statement in node.orelse]
    header = div(f"while {test}:")
    body = "".join(body)
    body = block(body)
    return div(header + body)
def render_for(node):
    target = render_expr(node.target)
    it = render_expr(node.iter)
    header = div(f"for {target} in {it}:")
    body = [render_statement(statement) for statement in node.body]
    body = dom.block("".join(body))

    block = div(header + body)

    parts = [block]

    if node.orelse:
        else_header = div("else:")
        else_body = [render_statement(statement) for statement in node.orelse]
        else_body = block("".join(else_body))
        else_block = div(else_header + else_body)
        parts.append(else_block)

    assert node.type_comment is None

    result = "".join(parts)
    return div(result)


def render_assert(node):
    # TODO: support assertion messages
    assert node.msg is None
    test = render_expr(node.test)
    return div(f"assert {test}")
def render_return(node):
    if node.value is not None:
        result = render_expr(node.value)
        return div(f"return {result}")
    else:
        return ""
def render_if(node):
    # manual processing
    # the Python AST represent elif branches as nested else: if
    if is_elif(node):
        #print("found elif")
        return render_elifs(node)
    test = render_expr(node.test)
    body = [render_statement(s) for s in node.body]

    if_header = div(f"if {test}:")
    body = block("".join(body))

    ifpart = div(if_header + body)

    parts = [ifpart]

    if node.orelse:
        else_header = f"<div>else:</div>"
        else_body = [render_statement(s) for s in node.orelse]
        else_body = "".join(else_body)
        else_body = f"<div style='margin-left: 30px'>{else_body}</div>"

        elsepart = f"<div>{else_header}{else_body}</div>"
        parts.append(elsepart)
    return "".join(parts)

# helpers for elifs
def is_elif(if_node):
    "return True if the if has an elif branch"
    # if/elif gets compiled to if/else{if}
    # if/elif/else gets compiled to if/else{if/else}
    if len(if_node.orelse) != 1:
        return False
    return isinstance(if_node.orelse[0], ast.If)

def elif_depth(if_node):
    "return the number of elif branches in an if"
    n = 0
    while is_elif(if_node):
        n += 1
        if_node = if_node.orelse[0]
    return n

def collect_branches(if_node):
    n = elif_depth(if_node)
    branches = []
    for i in range(n):
        condition = if_node.test
        body = if_node.body
        branches.append((condition, body))
        if_node = if_node.orelse[0]
    # final else (not if) branch
    if if_node.orelse:
        assert not isinstance(if_node.orelse[0], ast.If)
    # the last element is not a (cond, body) tuple
    branches.append(if_node.orelse)
    return branches

def render_elifs(if_node):
    branches = collect_branches(if_node)
    if_branch, *elif_branches, else_body = branches


    # if
    if_test, if_body = if_branch
    if_test = render_expr(if_test)
    if_header = f"<div>if {if_test}:</div>"

    if_body = [render_statement(s) for s in if_body]
    if_body = "".join(if_body)
    if_body = f"<div style='margin-left: 30px'>{if_body}</div>"

    ifpart = f"<div>{if_header}{if_body}</div>"

    parts = [ifpart]

    # elifs
    for (test, body) in elif_branches:
        test = render_expr(test)
        header = f"<div>elif {test}:</div>"
        body = [render_statement(s) for s in body]
        body = "".join(body)
        body = f"<div style='margin-left: 30px'>{body}</div>"
        part = f"<div>{header}{body}</div>"
        parts.append(part)

    # else
    if else_body:
        else_header = f"<div>else:</div>"
        else_body = [render_statement(s) for s in else_body]
        else_body = "".join(else_body)
        else_body = f"<div style='margin-left: 30px'>{else_body}</div>"
        elsepart = f"<div>{else_header}{else_body}</div>"
        parts.append(elsepart)

    return "".join(parts)
    #return ifpart



def render_pass(node):
    return "<div>pass</div>"
def render_try(node):
    body = [render_statement(statement) for statement in node.body]
    handlers = [render_except_handler(h) for h in node.handlers]
    header = f"<div>try:</div>"

    body = "".join(body)
    body = f"<div style='margin-left: 30px'>{body}</div>"

    parts = [header, body]

    for h in handlers:
        parts.append(h)

    if node.orelse:
        header = "<div>else:</div>"
        else_body = [render_statement(statement) for statement in node.orelse]
        else_body = "".join(else_body)
        else_body = f"<div style='margin-left: 30px'>{else_body}</div>"
        else_block = f"<div>{header}{else_body}</div>"
        parts.append(else_block)

    if node.finalbody:
        header = "<div>finally:</div>"
        final_body = [render_statement(statement) for statement in node.finalbody]
        final_body = "".join(final_body)
        final_body = f"<div style='margin-left: 30px'>{final_body}</div>"
        final_block = f"<div>{header}{final_body}</div>"
        parts.append(final_block)
    result = "".join(parts)
    return f"<div>{result}</div>"

def render_except_handler(node):
    error_type = render_expr(node.type)
    body = [render_statement(statement) for statement in node.body]
    header = f"<div>except {error_type} as {node.name}:</div>"
    body = "".join(body)
    body = f"<div style='margin-left: 30px'>{body}</div>"
    return f"<div>{header}{body}</div>"

def render_arguments(node):
    result = ", ".join([render_arg(arg) for arg in node.args])
    # TODO: support more cases
    assert len(node.posonlyargs) == 0
    assert len(node.kwonlyargs) == 0
    assert len(node.kw_defaults) == 0
    # TODO: implement in the static renderer (implemented in the dynamic renderer)
    if len(node.defaults) != 0:
        pass
    # flags (*args and **kwargs)
    # TODO: support this instead of ignoring it
    #assert node.vararg is None
    assert node.kwarg is None
    return result
def render_arg(arg):
    # TODO: support argument type annotations (implemented in the dynamic renderer)
    if arg.annotation is not None:
        pass
    assert arg.type_comment is None
    # text metadata (not needed in a no-text IDE)
    #print(arg.lineno)
    #print(arg.col_offset)
    #print(arg.end_lineno)
    #print(arg.end_col_offset)
    return arg.arg

def render_raise(node):
    # TODO: check if support is needed
    assert node.cause is None
    exc = render_expr(node.exc)
    return f"<div>raise {exc}</div>"


def render_expr_statement(node):
    return f"<div>{render_expr(node)}</div>"


"""
expression rendering
"""

def render_cmpop(node):
    if isinstance(node, ast.Eq):
        return "=="
    elif isinstance(node, ast.NotEq):
        return "!="
    elif isinstance(node, ast.Lt):
        return html.escape("<")
    elif isinstance(node, ast.LtE):
        return html.escape("<=")
    elif isinstance(node, ast.Gt):
        return html.escape(">")
    elif isinstance(node, ast.GtE):
        return html.escape(">=")
    elif isinstance(node, ast.Is):
        return "is"
    elif isinstance(node, ast.IsNot):
        return "is not"
    elif isinstance(node, ast.In):
        return "in"
    elif isinstance(node, ast.NotIn):
        return "not in"
    else:
        raise NotImplementedError("unknown comparison operator")


def render_unaryop(op):
    if isinstance(op, ast.USub):
        return "-"
    elif isinstance(op, ast.UAdd):
        return "+"
    elif isinstance(op, ast.Not):
        return "not "
    elif isinstance(op, ast.Invert):
        return "~"
    else:
        raise NotImplementedError("unknown unary operator")


def render_binaryop(op):
    if isinstance(op, ast.Add):
        return "+"
    elif isinstance(op, ast.Sub):
        return "-"
    elif isinstance(op, ast.Mult):
        return "*"
    elif isinstance(op, ast.Div):
        pass
    elif isinstance(op, ast.FloorDiv):
        pass
    elif isinstance(op, ast.Mod):
        pass
    elif isinstance(op, ast.Pow):
        pass
    elif isinstance(op, ast.LShift):
        pass
    elif isinstance(op, ast.RShift):
        pass
    elif isinstance(op, ast.BitOr):
        pass
    elif isinstance(op, ast.BitXor):
        pass
    elif isinstance(op, ast.BitAnd):
        pass
    elif isinstance(op, ast.MatMult):
        pass
    raise NotImplementedError(f"unknown binary operator: {op}")

def render_boolop(op):
    if isinstance(op, ast.And):
        return "and"
    elif isinstance(op, ast.Or):
        return "or"
    else:
        raise NotImplementedError(f"unknown boolean operator: {op}")


def render_keyword(node):
    value = render_expr(node.value)
    return f"{node.arg}={value}"

def render_expr(node):
    if isinstance(node, ast.Expr):
        # nested Expr wrapper
        return render_expr(node.value)
    elif isinstance(node, ast.BoolOp):
        op = render_boolop(node.op)
        values = [render_expr(v) for v in node.values]
        result = f" {op} ".join(values)
        return result
    #elif isinstance(node, ast.NamedExpr):
    #    return "<span style='color: red'>TODO</span> named"
    elif isinstance(node, ast.BinOp):
        left = render_expr(node.left)
        op = render_binaryop(node.op)
        right = render_expr(node.right)
        return f"{left} {op} {right}"
    elif isinstance(node, ast.UnaryOp):
        op = render_unaryop(node.op)
        value = render_expr(node.operand)
        return f"{op}{value}"
    #elif isinstance(node, ast.Lambda):
    #    return "<span style='color: red'>TODO</span> "
    elif isinstance(node, ast.IfExp):
        test = render_expr(node.test)
        if_expr = render_expr(node.body)
        else_expr = render_expr(node.orelse)
        return f"{if_expr} if {test} else {else_expr}"
    elif isinstance(node, ast.Dict):
        return "<span style='color: red'>TODO</span> "
    #elif isinstance(node, ast.Set):
    #    return "<span style='color: red'>TODO</span> "
    elif isinstance(node, ast.ListComp):
        expr = render_expr(node.elt)
        generators = [render_comprehension(g) for g in node.generators]
        generators = [f"for {g}" for g in generators]
        # TODO: test with more than 1 generator
        generators = " ".join(generators)
        return f"[{expr} {generators}]"
    #elif isinstance(node, ast.SetComp):
    #    return "<span style='color: red'>TODO</span> "
    #elif isinstance(node, ast.DictComp):
    #    return "<span style='color: red'>TODO</span> "
    #elif isinstance(node, ast.GeneratorExp):
    #    return "<span style='color: red'>TODO</span> "
    #elif isinstance(node, ast.Await):
    #    return "<span  style='color: red'>TODO</span> "
    #elif isinstance(node, ast.Yield):
    #    return "<span style='color: red'>TODO</span> "
    #elif isinstance(node, ast.YieldFrom):
    #    return "<span style='color: red'>TODO</span> "
    elif isinstance(node, ast.Compare):
        # in python, comparisons can be complex sequences, like:
        # 1 < x < y < 6
        # 1 is called left
        # the operators are [<, <, <]
        # and the comparators are [x, y, 6]
        left = render_expr(node.left)
        operators = [render_cmpop(op) for op in node.ops]
        comparators = [render_expr(cmp) for cmp in node.comparators]
        items = [left]
        for op, cmp in zip(operators, comparators):
            items.append(op)
            items.append(cmp)
        return " ".join(items)
    elif isinstance(node, ast.Call):
        # positional args
        args = [render_expr(arg) for arg in node.args]
        # kwargs
        args.extend([render_keyword(kw) for kw in node.keywords])
        args = ", ".join(args)
        func = render_expr(node.func)
        return f"{func}({args})"
    elif isinstance(node, ast.FormattedValue):
        # TODO: support other values
        # 97: !a, ascii
        # 114: !r, repr() formatting
        # 115: !s, string formatting
        assert node.conversion == -1
        # nested JoinedStr or None
        assert node.format_spec is None
        return render_expr(node.value)
    #elif isinstance(node, ast.Interpolation):
    #    return "<span style='color: red'>TODO</span> "
    elif isinstance(node, ast.JoinedStr):
        parts = []
        for e in node.values:
            if isinstance(e, ast.FormattedValue):
                parts.append("{" + render_expr(e) + "}")
            else:
                # skip render_expr to avoid representing substrings with quotes
                assert isinstance(e, ast.Constant)
                assert isinstance(e.value, str)
                parts.append(e.value)
        string_result = "".join(parts)
        return html.escape(repr(string_result))
    #elif isinstance(node, ast.TemplateStr):
    #    return "<span style='color: red'>TODO</span> "
    elif isinstance(node, ast.Constant):
        assert node.kind is None
        # TODO: test carefully
        # there could be other places where html can get by mistake
        # html.escape() means we can't have html in the string literals of the edited source code
        if isinstance(node.value, str):
            return repr(html.escape(node.value))
        else:
            return repr(node.value)
    elif isinstance(node, ast.Attribute):
        obj = render_expr(node.value)
        attr = node.attr
        #print(node.ctx)
        #assert isinstance(node.ctx, ast.Load)
        return f"{obj}.{attr}"
    elif isinstance(node, ast.Subscript):
        indexed = render_expr(node.value)
        index = render_expr(node.slice)
        # node.ctx is either ast.Load or ast.Store
        # Store if the subscript is in a left side of an assignment
        # Load if the subscript is in an expression to evaluate
        return f"{indexed}[{index}]"
    elif isinstance(node, ast.Starred):
        #print(node.ctx)
        expr = render_expr(node.value)
        return f"*{expr}"
    elif isinstance(node, ast.Name):
        # the context can be Store (variable to store in, in assignment), Load (value to load), or Del (in del statements)
        # never used here for now
        #assert isinstance(node.ctx, ast.Store)
        return node.id
    elif isinstance(node, ast.List):
        elts = [render_expr(expr) for expr in node.elts]
        assert isinstance(node.ctx, ast.Load)
        elts = ", ".join(elts)
        return f"[{elts}]"
    elif isinstance(node, ast.Tuple):
        #print(node.ctx)
        elts = [render_expr(elt) for elt in node.elts]
        if len(elts) == 0:
            return "(,)"
        if len(elts) == 1:
            return f"({elts[0]},)"
        elts = ", ".join(elts)
        return f"({elts})"
    elif isinstance(node, ast.Slice):
        return "<span style='color: red'>TODO</span> "
    else:
        raise NotImplementedError(type(node))

def render_comprehension(node):
    target = render_expr(node.target)
    it = render_expr(node.iter)
    # TODO: support async
    assert node.is_async == 0
    # TODO: support conditions in comprehensions
    assert len(node.ifs) == 0
    #ifs = [render_expr(x) for x in node.ifs]
    #print(ifs)
    return f"{target} in {it}"



class CodeWindow:
    def __init__(self, filepath, load=True):
        self.path = filepath
        with open(self.path) as f:
            self.source = f.read()
        self.module_name = os.path.basename(filepath)
        self.module = ModuleType(self.module_name)
        
        # TODO: use correct namespaces for packages to allow non-root imports
        sys.modules[filepath] = self.module

        # TODO: use the new renderer
        self.tree = ast.parse(self.source)
        self.html = render_module(self.tree)

        # the API cannot have a CodeWindow as an attribute
        # the constructor cannot have parameters
        # closures are the only way to store self as the context of the API
        # so this class has to be defined here
        class CodeWindowAPI:
            js_init = """
            document.body.addEventListener("keydown", (e) => {
            pywebview.api.keydown(e.key)
            })
            """

            def keydown(_, key):
                if key == "r":
                    bytecode = compile(self.tree, "<ast>", "exec")
                    exec(bytecode, self.module.__dict__)
                    #def f():
                    #    print(os.getpid())
                    #    print(os.getppid())
                    #    exec(bytecode, self.module.__dict__)
                    #encoded = marshal.dumps(bytecode)
                    #p = multiprocessing.Process(target=run_code, args=(encoded, self.module.__dict__))
                    #p.start()
                    #p.join()
                    #exec(bytecode, self.module.__dict__)
                elif key == "s":
                    # start an REPL in the current module
                    code.InteractiveConsole(locals=self.module.__dict__).interact()
                print(key)

        self.api = CodeWindowAPI()
        self.window = webview.create_window(self.module_name, html=self.html, js_api=self.api)

        if load:
            self.load()


    # must be called after webview.start()
    # (load_css, evaluate_js and dom manipulation cannot be done before opening the window)
    def load(self):
        # inject css into the window and load the html
        for name in ["style.css", "syntax.css"]:
            css = resources.files("untext.css").joinpath(name).read_text()
            #with open(name) as f:
            #    css = f.read()
            self.window.load_css(css)
        self.window.evaluate_js(self.api.js_init)
        self.root = self.window.dom.create_element("<div id='root'></div>")
        # will fail until the new renderer is complete
        statement.render_module(self.root, self.tree)




def main():
    # open files listed in sys.argv
    windows = []
    for path in sys.argv[1:]:
        print(path)
        if not os.path.exists(path):
            # create the file
            with open(path, "x"):
                pass
        windows.append(CodeWindow(path, load=False))

    def on_load():
        for win in windows:
            win.load()

    if not windows:
        print("Usage: untext <file1> <file2>")
        sys.exit(0)

    print("starting")
    webview.start(on_load)


if __name__ == "__main__":
    main()




"""
statement rendering
"""

# types
from webview.dom.dom import DOM
from webview.dom.element import Element
import ast

from .dom import register, div, block, add_node, add, add_text
from . import expression


def render(parent: Element, node: ast.stmt):
    match type(node):
        case ast.FunctionDef:
            return render_funcdef(parent, node)
        case ast.AsyncFunctionDef:
            raise NotImplementedError("statement.render() not implemented for ast.AsyncFunctionDef")
        case ast.ClassDef:
            return render_classdef(parent, node)
        case ast.Return:
            return render_return(parent, node)

        case ast.Delete:
            raise NotImplementedError("statement.render() not implemented for ast.Delete")
        case ast.Assign:
            return render_assign(parent, node)
        # 3.12+ feature
        # TODO: ignore it until pypy reaches 3.12 or stop supporting pypy
        # case ast.TypeAlias:
        #     raise NotImplementedError("statement.render() not implemented for ast.TypeAlias")
        case ast.AugAssign:
            return render_augassign(parent, node)
        case ast.AnnAssign:
            raise NotImplementedError("statement.render() not implemented for ast.AnnAssign")

        case ast.For:
            return render_for(parent, node)
        case ast.AsyncFor:
            raise NotImplementedError("statement.render() not implemented for ast.AsyncFor")
        case ast.While:
            return render_while(parent, node)
        case ast.If:
            return render_if(parent, node)
        case ast.With:
            return render_with(parent, node)
        case ast.AsyncWith:
            raise NotImplementedError("statement.render() not implemented for ast.AsyncWith")

        case ast.Match:
            return render_match(parent, node)

        case ast.Raise:
            return render_raise(parent, node)
        case ast.Try:
            raise NotImplementedError("statement.render() not implemented for ast.Try")
        case ast.TryStar:
            raise NotImplementedError("statement.render() not implemented for ast.TryStar")
        case ast.Assert:
            return render_assert(parent, node)

        case ast.Import:
            return render_import(parent, node)
        case ast.ImportFrom:
            return render_importfrom(parent, node)

        case ast.Global:
            raise NotImplementedError("statement.render() not implemented for ast.Global")
        case ast.Nonlocal:
            return render_nonlocal(parent, node)
        case ast.Expr:
            return expression.render(parent, node.value)
        case ast.Pass:
            return render_pass(parent, node)
        case ast.Break:
            raise NotImplementedError("statement.render() not implemented for ast.Break")
        case ast.Continue:
            raise NotImplementedError("statement.render() not implemented for ast.Continue")

        case default:
            raise ValueError(f"Unexpected ast statement type: {type(node)}")



# TODO: find a place to put this function (main.py ?)
# modules are not statements, they are top-level programs
def render_module(parent: Element, node: ast.Module) -> Element:
    elt = add_node(parent, node)
    for stmt in node.body:
        render(elt, stmt)
    return elt


"""
AST statement node rendering
"""

def render_match(parent: Element, node: ast.Match):
    elt = add_node(parent, node)
    header = add(elt)
    colon_suffix = add(header, "colon-suffix row")
    match_prefix = add(colon_suffix, "match-prefix row gap")
    matched_value = expression.render(match_prefix, node.subject)

    cases = add(elt, "block")
    for case in node.cases:
        render_case(cases, case)

    return elt

# part of render_match
def render_case(parent: Element, node: ast.match_case):
    # TODO: support more cases
    assert node.guard is None
    elt = add_node(parent, node)
    header = add(elt, "row colon-suffix")
    content = add(header, "row gap case-prefix")
    # turns out match patterns are not actually expressions,
    # despite having the same syntax
    render_pattern(content, node.pattern)

    body = add(elt, "block")
    for stmt in node.body:
        render(body, stmt)

# TODO: implement missing cases
def render_pattern(parent: Element, node: ast.pattern):
    match type(node):
        case ast.MatchValue:
            return render_match_value(parent, node)
        case ast.MatchSingleton:
            return render_match_singleton(parent, node)
        case ast.MatchSequence:
            return render_match_sequence(parent, node)
        case ast.MatchMapping:
            return render_match_mapping(parent, node)
        case ast.MatchClass:
            return render_match_class(parent, node)

        case ast.MatchStar:
            return render_match_star(parent, node)

        case ast.MatchAs:
            return render_match_as(parent, node)
        case ast.MatchOr:
            return render_match_or(parent, node)

        case default:
            raise ValueError(f"Unexpected match pattern type: {type(node)}")


def render_match_value(parent: Element, node: ast.MatchValue):
    elt = add_node(parent, node)
    expression.render(elt, node.value)

def render_match_as(parent: Element, node: ast.MatchAs):
    # TODO: support other cases
    # case [x] as y:
    # case default:     <--- the only kind used

    # specific to case default:
    assert node.pattern is None
    assert node.name == "default"
    elt = add_node(parent, node)
    add_text(elt, node.name)
    return elt





def render_raise(parent: Element, node: ast.Raise) -> Element:
    # TODO: check if support for this attribute is needed
    assert node.cause is None
    elt = add_node(parent, node, "raise-prefix row gap")
    expression.render(elt, node.exc)
    return elt


def render_assert(parent: Element, node: ast.Assert) -> Element:
    # TODO: support assertion messages
    assert node.msg is None
    elt = add_node(parent, node, "assert-prefix gap row")
    expression.render(elt, node.test)
    return elt

def render_import(parent: Element, node: ast.Import) -> Element:
    elt = add_node(parent, node, "import import-prefix row")
    aliases = add(elt, "aliases row comma-sep")
    for name in node.names:
        #comma_separated_item = add(aliases, "row")
        render_alias(add(aliases, "row gap"), name)
    return elt

# sub-part of import and importfrom nodes
def render_alias(parent: Element, node: ast.alias) -> Element:
    elt = add_node(parent, node, "alias row")
    if node.asname is not None:
        # create an alias (div (div name) "as" (div asname))
        alias = add(elt, "import-alias as-sep row gap")
        name = add_text(add(alias, "row gap"), node.name)
        asname = add_text(add(alias, "row gap"), node.asname)
    else:
        # TODO: find a class name for "unaliased import"
        alias = add(elt)
        name = add_text(alias, node.name)
    return elt


def render_importfrom(parent: Element, node: ast.ImportFrom) -> Element:
    elt = add_node(parent, node, "importfrom row gap")

    # prefixed items need .row and .gap to space their prefix and content
    from_prefixed = add(elt, "from-prefix row gap")
    from_content = add(from_prefixed, "row")
    relative_level = add_text(from_content, "." * node.level)
    from_field = add_text(from_content, node.module)

    import_prefixed = add(elt, "import-prefix row gap")
    aliases = add(import_prefixed, "aliases row comma-sep")
    for name in node.names:
        render_alias(add(aliases, "row gap"), name)
    return elt


def render_funcdef(parent: Element, node: ast.FunctionDef) -> Element:
    # 3.12+ feature
    # instead, see: type_comment
    # TODO: wait for pypy to reach 3.12 or explicitely stop supporting pypy
    #assert len(node.type_params) == 0
    assert len(node.decorator_list) == 0
    assert node.type_comment is None
    elt = add_node(parent, node, "funcdef")

    header = add(elt, "row colon-suffix")
    spaced_signature = add(header, "row gap return-type-arrow-sep")
    left = add(spaced_signature, "row")
    funcname = add(left, "row gap def-prefix", node.name)
    params = add(left, "parens row")
    if node.returns is not None:
        right = add(spaced_signature, "row gap")
        expression.render(right, node.returns)
    #params_content = add(params, "comma-sep row gap")
    render_parameters(params, node.args)
    body = add(elt, "block")

    for stmt in node.body:
        render(body, stmt)
    return elt


# sub-part of render_funcdef
def render_parameters(parent: Element, node: ast.arguments) -> Element:
    # TODO: support more cases
    assert len(node.posonlyargs) == 0
    assert len(node.kwonlyargs) == 0
    assert len(node.kw_defaults) == 0
    # default values are for the last parameters
    # need to match argument names to default values with indices
    default_padding = len(node.args) - len(node.defaults)
    # flags (*args and **kwargs)
    assert node.vararg is None
    assert node.kwarg is None
    elt = add(parent, "comma-sep row")
    for i, param in enumerate(node.args):
        comma_separated_item = add(elt, "row gap")
        if i >= default_padding:
            param_elt = add(comma_separated_item, "equal-sep row gap")
            render_param(param_elt, param)
            expression.render(add(param_elt, "row gap"), node.defaults[i - default_padding])
        else:
            render_param(comma_separated_item, param)
    #result = ", ".join([render_arg(arg) for arg in node.args])
    #return result
    return elt

# sub-part of render_parameters
def render_param(parent: Element, node: ast.arg) -> Element:
    assert node.type_comment is None
    # text metadata (not needed in a no-text IDE)
    #print(arg.lineno)
    #print(arg.col_offset)
    #print(arg.end_lineno)
    #print(arg.end_col_offset)

    # comma-separated items must be inlined,
    # so that the comma is on the same line
    elt = add_node(parent, node, "row")
    if node.annotation is None:
        add(elt, text=node.arg)
    else:
        typed_group = add(elt, "row gap")
        name = add(typed_group, "row colon-suffix", text=node.arg)
        expression.render(typed_group, node.annotation)
    return elt

def render_classdef(parent: Element, node: ast.ClassDef) -> Element:
    # TODO: support decorators
    # TODO: support type_params
    assert not node.decorator_list
    elt = add_node(parent, node)
    # header
    colon_suffixed = add(elt, "row colon-suffix")
    class_prefixed = add(colon_suffixed, "class-prefix row gap")
    header_content = add(class_prefixed, "row")
    name = add(header_content, text=node.name)
    if node.keywords or node.bases:
        paren_wrapped = add(header_content, "parens row")
        arguments = add(paren_wrapped, "comma-sep row")
        for kwarg in node.keywords:
            expression.render_keyword_arg(add(arguments, "row gap"), kwarg)
        for base in node.bases:
            expression.render(add(arguments, "row gap"), base)

    body = add(elt, "block")
    for stmt in node.body:
        render(body, stmt)
    return elt

def render_return(parent: Element, node: ast.Return) -> Element:
    elt = add_node(parent, node, "return-prefix row gap")
    if node.value is not None:
        expression.render(elt, node.value)
    return elt

def render_assign(parent: Element, node: ast.Assign) -> Element:
    assert node.type_comment is None
    # TODO: support multiple targets
    # example:
    # a = b = 1
    assert len(node.targets) == 1
    elt = add_node(parent, node, "row equal-sep gap")
    variables = add(elt, "row gap")
    #for t in node.targets:
    expression.render(variables, node.targets[0])
    value = add(elt, "row gap")
    expression.render(value, node.value)
    return
    #if len(targets) == 1:
    #    target = targets[0]
    #    return div(f"{target} = {value}")
    #else:
    #    return div(f"{tuple(targets)} = {value}")
    return elt


# TODO: fix: the operator must be displayed
# format: <node> <op>= <node>
# <node><gap><op>=<gap><node>
# (<node> (<op>=) <node>)
def render_augassign(parent: Element, node: ast.AugAssign) -> Element:
    elt = add_node(parent, node, "row gap")
    target = expression.render(add(elt), node.target)
    # hack: add an empty div to force the = separator to render
    # TODO?: add a .equal-suffix css class
    assign_operator = add(elt, "equal-sep row")
    add(assign_operator, text=expression.read_binaryop(node.op))
    add(assign_operator, "row")
    #op = expression.read_binaryop(node.op)
    #elt.classes.append(f"{op}-sep")
    val = expression.render(add(elt), node.value)
    return elt



def render_for(parent: Element, node: ast.For) -> Element:
    assert node.type_comment is None
    elt = add_node(parent, node)
    header = add(elt, "row colon-suffix")
    prefixed = add(header, "for-prefix in-sep row gap")
    # separators like in-sep need an additional div (add(elt)) to add the separator as a suffix of this wrapper div
    target = expression.render(prefixed, node.target)
    iterator = expression.render(add(prefixed, "row gap"), node.iter)

    # TODO: WIP, debug later
    body = add(elt, "block")
    for stmt in node.body:
        render(body, stmt)

    # TODO: process for: else: blocks
    # (add more headers after the main body)
    assert not node.orelse
    #parts = [block]
    #if node.orelse:
    #    else_header = div("else:")
    #    else_body = [render_statement(statement) for statement in node.orelse]
    #    else_body = block("".join(else_body))
    #    else_block = div(else_header + else_body)
    #    parts.append(else_block)

    #result = "".join(parts)
    #return div(result)


    return elt



def render_while(parent: Element, node: ast.While) -> Element:
    elt = add_node(parent, node)
    header = add(elt, "colon-suffix row")
    header_content = add(header, "while-prefix row gap")
    test = expression.render(header_content, node.test)
    body = add(elt, "block")
    for stmt in node.body:
        render(body, stmt)
    assert not node.orelse
    # if node.orelse:
    #else_body = [render_statement(statement) for statement in node.orelse]
    return elt

def render_if(parent: Element, node: ast.If) -> Element:
    elt = add_node(parent, node)
    if is_elif(node):
        return render_elifs(elt, node)
    header = add(elt, "row colon-suffix")
    header_content = add(header, "row gap if-prefix")
    expression.render(header_content, node.test)
    block = add(elt, "block")
    for stmt in node.body:
        render(block, stmt)
    if node.orelse:
        else_header = add(elt, "row colon-suffix else-prefix")
        else_block = add(elt, "block")
        for stmt in node.orelse:
            render(else_block, stmt)

    return elt


# helpers for elifs
# internally, Python represents elifs as nested else: if:
def is_elif(if_node: ast.If):
    "return True if the if has an elif branch"
    # if/elif gets compiled to if/else{if}
    # if/elif/else gets compiled to if/else{if/else}
    if len(if_node.orelse) != 1:
        return False
    return isinstance(if_node.orelse[0], ast.If)

def elif_depth(if_node: ast.If):
    "return the number of elif branches in an if"
    n = 0
    while is_elif(if_node):
        n += 1
        if_node = if_node.orelse[0]
    return n

def collect_branches(if_node: ast.If):
    n = elif_depth(if_node)
    branches = []
    for i in range(n):
        condition = if_node.test
        body = if_node.body
        branches.append((condition, body))
        if_node = if_node.orelse[0]
    # final else (not if) branch
    if if_node.orelse:
        assert not isinstance(if_node.orelse[0], ast.If)
    # the last element is not a (cond, body) tuple
    branches.append(if_node.orelse)
    return branches

def render_elifs(elt: Element, if_node: ast.If):
    branches = collect_branches(if_node)
    if_branch, *elif_branches, else_body = branches


    # if
    if_test, if_body = if_branch
    if_elt = add(elt)
    header = add(if_elt, "row colon-suffix")
    header_content = add(header, "row gap if-prefix")
    _if_test = expression.render(header_content, if_test)

    body = add(if_elt, "block")
    for stmt in if_body:
        render(body, stmt)

    # elifs
    for (test, ast_body) in elif_branches:
        elif_elt = add(elt)
        header = add(elif_elt, "row colon-suffix")
        header_content = add(header, "row gap elif-prefix")
        _test = expression.render(header_content, test)
        body = add(elif_elt, "block")
        for stmt in ast_body:
            render(body, stmt)

    # else
    if else_body:
        else_elt = add(elt)
        header = add(else_elt, "row colon-suffix")
        header_content = add(header, "row gap else-prefix")
        body = add(else_elt, "block")
        for stmt in else_body:
            render(body, stmt)

    return elt






def render_with(parent: Element, node: ast.With) -> Element:
    assert node.type_comment is None
    elt = add_node(parent, node)
    # this breaks if newlines are added for clarity:
    #header = elt.append("""<div class="with-prefix colon-suffix row gap"></div>""")
    header = add(elt)
    colon_suffixed = add(header, "row colon-suffix")
    header_content = add(colon_suffixed, "with-prefix row gap")
    #print(header)
    body = add(elt, "block")
    for item in node.items:
        render_withitem(header_content, item)
    for stmt in node.body:
        render(body, stmt)
    return elt

def render_withitem(parent: Element, node: ast.withitem) -> Element:
    elt = add_node(parent, node)
    if node.optional_vars:
        # add "<expr> as <name>"
        named = add(elt, "as-sep row gap")
        expr = expression.render(add(named, "row gap"), node.context_expr)
        name = expression.render(add(named, "row gap"), node.optional_vars)
    else:
        # just add <expr>
        unnamed = add(elt)
        expr = expression.render(unnamed, node.context_expr)
    return elt


def render_nonlocal(parent: Element, node: ast.Nonlocal):
    elt = add_node(parent, node, "row gap nonlocal-prefix")
    names = add(elt, "row comma-sep")
    for name in node.names:
        add(names, "row gap", name)


def render_pass(parent: Element, node: ast.Pass) -> Element:
    elt = add_node(parent, node, text="pass")
    return elt

